USE MeineTestDB
GO

/*
	Beispiel Hierarchy 
	================

                                 Martin-100
                                    |
                   +----------------+---------------+
                   |                |               |
                Julia-121          Peter-122        Maria-123
                   |                |               |
              +----+----+           |               |
              |         |           |               |
         Rosi-233  Wanda-234  Martina-235        Kevin-236
              |
         +----+----+
         |         |
    Richard-310   Uli-311
*/

-- Erzeugen der hierarchischen Tabelle mit einem Index fr die Abfrage in die Tiefe der Hierarchy
CREATE TABLE Mitarbeiter
(
   NodeId			hierarchyid PRIMARY KEY CLUSTERED,
   NodeLevel		AS NodeId.GetLevel(),
   MitarbeiterID    int UNIQUE NOT NULL,
   MitarbeiterName	varchar(20) NOT NULL,
   Titel			varchar(20) NULL
)
GO

-- Erzeugen der Wurzel
INSERT INTO Mitarbeiter
  (NodeId, MitarbeiterID, MitarbeiterName, Titel)
 VALUES
  (hierarchyid::GetRoot(), 100, 'Martin', 'Marketing Manager') ;

GO

SELECT * FROM Mitarbeiter
GO

-- Einfgen von Julia als erstes Kind von Martin
DECLARE @Manager hierarchyid 

SELECT @Manager = NodeId
 FROM Mitarbeiter
 WHERE MitarbeiterID = 100

INSERT INTO Mitarbeiter
  (NodeId, MitarbeiterID, MitarbeiterName, Titel)
 VALUES
  (@Manager.GetDescendant(NULL, NULL), 121, 'Julia', 'Marketing Spezialist')

GO

-- Alternativ geht es auch in der single-statement Version:
INSERT INTO Mitarbeiter
  (NodeId, MitarbeiterID, MitarbeiterName, Titel)
 VALUES (
  (SELECT NodeId
    FROM Mitarbeiter
    WHERE MitarbeiterID = 100).GetDescendant(NULL, NULL),
  121,
  'Julia',
  'Marketing Spezialist')

GO

SELECT NodeId.ToString() AS NodeIdPath, *
 FROM Mitarbeiter
GO

-- Erzeugen einer stored procedure zum einfacheren Einfgen
CREATE PROC uspAddMitarbeiter(
 @ManagerID int,
 @MitarbeiterID int,
 @MitarbeiterName varchar(20),
 @Titel varchar(20)) 
AS 
BEGIN

 DECLARE @ManagerNodeId hierarchyid
 DECLARE @LetztesManagerKind hierarchyid
 DECLARE @NeuerMitarbeiterNodeId hierarchyid

 -- HierarchyID des gewnschten Managers der durch @ManagerID bergeben wurde
 SELECT @ManagerNodeId = NodeId 
  FROM  Mitarbeiter 
  WHERE MitarbeiterID = @ManagerID

 SET TRANSACTION ISOLATION LEVEL SERIALIZABLE

 BEGIN TRANSACTION

   -- Ermitteln der HierarchyID des letzten existierenden Kindes unterhalb des Managers
   SELECT @LetztesManagerKind = MAX(NodeId)
    FROM  Mitarbeiter
    WHERE NodeId.GetAncestor(1) = @ManagerNodeId

   -- Zuordnen einer neuen HierarchyID am Ende aller existierenden Geschwister
   SELECT @NeuerMitarbeiterNodeId =
    @ManagerNodeId.GetDescendant(@LetztesManagerKind, NULL)

   -- Zeile einfgen
   INSERT INTO Mitarbeiter
     (NodeId, MitarbeiterID, MitarbeiterName, Titel)
    VALUES
     (@NeuerMitarbeiterNodeId, @MitarbeiterID, @MitarbeiterName, @Titel)

 COMMIT

END

GO

-- Die noch fehlenden Mitarbeiter einfgen
EXEC uspAddMitarbeiter 100, 122, 'Peter', 'Marketing Spezialist'
EXEC uspAddMitarbeiter 100, 123, 'Maria', 'Marketing Spezialist'
EXEC uspAddMitarbeiter 121, 233, 'Rosi', 'Marketing Assistent'
EXEC uspAddMitarbeiter 121, 234, 'Wanda', 'Business Assistent'
EXEC uspAddMitarbeiter 122, 235, 'Martina', 'Marketing Assistent'
EXEC uspAddMitarbeiter 123, 236, 'Kevin', 'Marketing Intern'
EXEC uspAddMitarbeiter 233, 310, 'Richard', 'Business Intern'
EXEC uspAddMitarbeiter 233, 311, 'Uli', 'Business Intern'

SELECT NodeId.ToString() AS NodeIdPath, * 
 FROM Mitarbeiter
 ORDER BY NodeLevel, NodeId

GO

-- Erzeugen einer UDF fr die Rckgabe des kompletten Pfades fr einen bestimmten Knoten
CREATE FUNCTION dbo.fnGetKomplettenPfad(@StartNodeId hierarchyid) 
 RETURNS varchar(max) 
AS 
  BEGIN
    DECLARE @StartLevelTiefe smallint
    DECLARE @LevelZaehler smallint
    DECLARE @Pfad varchar(max)
    DECLARE @ParentMitarbeiterName varchar(max)
 
    -- Los geht es mit dem angegebenen Knoten
    SELECT @StartLevelTiefe = NodeId.GetLevel(),
           @Pfad = MitarbeiterName
     FROM  Mitarbeiter
     WHERE NodeId = @StartNodeId
 
    -- Durch alle Vorfahren iterieren
    SET @LevelZaehler = 0
    WHILE @LevelZaehler < @StartLevelTiefe BEGIN
 
       SET @LevelZaehler = @LevelZaehler + 1
 
       SELECT @ParentMitarbeiterName = MitarbeiterName
        FROM  Mitarbeiter
        WHERE NodeId = (
              SELECT NodeId.GetAncestor(@LevelZaehler)
               FROM Mitarbeiter
               WHERE NodeId = @StartNodeId)
 
       -- Voranstellen des Vorfahren an den Pfad
       SET @Pfad = @ParentMitarbeiterName + ' > ' + @Pfad
 
    END
 
    RETURN(@Pfad)
   END 
GO

SELECT
  NodeId,
  NodeId.ToString() AS Pfad,
  dbo.fnGetKomplettenPfad(NodeId) AS LesbarerPfad,
  MitarbeiterName
 FROM
  Mitarbeiter 
 ORDER BY
  NodeLevel, NodeId
GO

-- Erzeugen eines Index fr die Abfrage in die Breite der Hierarchy
CREATE UNIQUE INDEX IX_MitarbeiterBreite
 ON Mitarbeiter(NodeLevel, NodeId)

GO

-- IsDescendantOf Methode: Abfrage des Baumes unterhalb von Julia
DECLARE @JuliaNodeId hierarchyid

SELECT @JuliaNodeId = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 121

SELECT NodeId.ToString() AS NodeIdPath, *
 FROM  Mitarbeiter
 WHERE NodeId.IsDescendantOf(@JuliaNodeId) = 1
 ORDER BY NodeLevel, NodeId

-- GetAncestor Methode: Abfrage der direkten Kinder von Julia (1 level tiefer)
SELECT NodeId.ToString() AS NodeIdPath, *
 FROM  Mitarbeiter
 WHERE NodeId.GetAncestor(1) = 
   (SELECT NodeId
     FROM  Mitarbeiter
     WHERE MitarbeiterID = 121)
 ORDER BY NodeLevel, NodeId

-- GetAncestor Methode: Abfrage von Martins Enkeln (2 levels tiefer)
SELECT NodeId.ToString() AS NodeIdPath, *
 FROM  Mitarbeiter
 WHERE NodeId.GetAncestor(2) = 
   (SELECT NodeId
     FROM  Mitarbeiter
     WHERE MitarbeiterID = 100)
 ORDER BY NodeLevel, NodeId
GO

-- GetRoot Methode: Abfrage der Wurzel (Martin)
SELECT NodeId.ToString() AS NodeIdPath, *
 FROM  Mitarbeiter
 WHERE NodeId = hierarchyid::GetRoot()
GO

-- GetReparentedValue Methode: Wanda reported nun an Maria und nicht mehr an Julia
DECLARE @MitarbeiterZumVerschieben hierarchyid
DECLARE @ParentAlt hierarchyid
DECLARE @ParentNeu hierarchyid

SELECT @MitarbeiterZumVerschieben = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 234 -- Wanda

SELECT @ParentAlt = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 121 -- Julia

SELECT @ParentNeu = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 123 -- Maria

UPDATE Mitarbeiter
 SET   NodeId = @MitarbeiterZumVerschieben.GetReparentedValue(@ParentAlt, @ParentNeu)
 WHERE NodeId = @MitarbeiterZumVerschieben
GO

SELECT
  NodeId,
  NodeId.ToString() AS NodeIdPath,
  NodeLevel,
  dbo.fnGetKomplettenPfad(NodeId) AS NodeIdDisplayPath,
  MitarbeiterName
 FROM
  Mitarbeiter 
 ORDER BY
  NodeLevel, NodeId
GO

-- Verschieben wir jetzt mal den kompletten Baum unterhalb von Julia an einen neuen Ort unterhalb von Kevin
DECLARE @ParentAlt hierarchyid
DECLARE @ParentNeu hierarchyid

SELECT @ParentAlt = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 121 -- Julia

SELECT @ParentNeu = NodeId
 FROM  Mitarbeiter
 WHERE MitarbeiterID = 236 -- Kevin

UPDATE Mitarbeiter
 SET   NodeId = NodeId.GetReparentedValue(@ParentAlt, @ParentNeu) 
 WHERE NodeId.IsDescendantOf(@ParentAlt) = 1
       AND MitarbeiterID <> 121 -- Julia selbst wird nicht mit verschoben
GO

SELECT
  NodeId,
  NodeId.ToString() AS NodeIdPath,
  NodeLevel,
  dbo.fnGetKomplettenPfad(NodeId) AS NodeIdDisplayPath,
  MitarbeiterName
 FROM
  Mitarbeiter
 ORDER BY
  NodeLevel, NodeId
GO

-- Parse Methode: (das Gegenstck zu ToString)
SELECT hierarchyid::Parse('/2/1/1/') AS NodeId
GO


/*
	GetDescendant() verstehen im Detail

                  Wurzel
                    |
       +------+-----+--+------+
       |      |        |      |
    Kind4   Kind1    Kind3  Kind2
              |
         +----+----+
         |         | 
       Enkel1   Enkel2
*/

DECLARE @Wurzel hierarchyid
DECLARE @Kind1 hierarchyid
DECLARE @Kind2 hierarchyid
DECLARE @Kind3 hierarchyid
DECLARE @Kind4 hierarchyid
DECLARE @Enkel1 hierarchyid
DECLARE @Enkel2 hierarchyid

SET @Wurzel = hierarchyid::GetRoot()
SET @Kind1 = @Wurzel.GetDescendant(NULL, NULL)
SET @Kind2 = @Wurzel.GetDescendant(@Kind1, NULL)
SET @Kind3 = @Wurzel.GetDescendant(@Kind1, @Kind2)
SET @Kind4 = @Wurzel.GetDescendant(NULL, @Kind1)
SET @Enkel1 = @Kind1.GetDescendant(NULL, NULL)
SET @Enkel2 = @Kind1.GetDescendant(@Enkel1, NULL)

SELECT
@Wurzel AS Root,
 @Kind1 AS Kind1,
 @Kind2 AS Kind2,
 @Kind3 AS Kind3,
 @Kind4 AS Kind4,
 @Enkel1 AS Enkel1,
 @Enkel2 AS Enkel2

SELECT
 @Wurzel.ToString() AS Root,
 @Kind1.ToString() AS Kind1,
 @Kind2.ToString() AS Kind2,
 @Kind3.ToString() AS Kind3,
 @Kind4.ToString() AS Kind4,
 @Enkel1.ToString() AS Enkel1,
 @Enkel2.ToString() AS Enkel2
GO

/*
	C L E A N U P - jetzt wird aufgerumt
*/

DROP FUNCTION dbo.fnGetKomplettenPfad
DROP PROC uspAddMitarbeiter
DROP TABLE Mitarbeiter
GO